#
#  Copyright 2001 Nathaniel E. Bray/Deekoo L.
#
#  This product includes software developed by Sampo Kellomaki (sampo@iki.fi)
#
#  See LICENSE for details.
#
package Yeemp::Channel;
use Socket;
use Net::SSLeay qw(die_now die_if_ssl_error);
require Exporter;


sub verify_callback()
{
  local $ok=$_[0];
  local $x509_ctx=$_[1];
  local $cert=Net::SSLeay::X509_STORE_CTX_get_current_cert($x509_ctx);
  local $name=Net::SSLeay::X509_NAME_oneline(Net::SSLeay::X509_get_subject_name($cert));

  if ($ok) {
    $verify_callback_ok{$target_domain}=1;
    return 1;
  }
}

sub verify_callback_newcert()
{
  local $ok=$_[0];
  local $x509_ctx=$_[1];
  local $cert=Net::SSLeay::X509_STORE_CTX_get_current_cert($x509_ctx);
  local $name=Net::SSLeay::X509_NAME_oneline(Net::SSLeay::X509_get_subject_name($cert));

  if ($ok) {
    $verify_callback_ok{$target_domain}=1;
    return 1;
  }
  if (exists($need_cert{$target_domain}) &&
    (!$Yeemp::Config::noninteractive)) {
    &Yeemp::Client::validate_unknown_cert($target_domain,
      Net::SSLeay::PEM_get_string_X509($cert));
    delete($need_cert{$target_domain});
  }
  return 1;
}

sub initialize_openssl()
{
  Net::SSLeay::load_error_strings();
  Net::SSLeay::ERR_load_crypto_strings();
  Net::SSLeay::SSLeay_add_ssl_algorithms();
  Net::SSLeay::randomize();
  $openssl_initialized=1;
}

sub channel_open($$)
{
  # Args: 
  # 0: SSLhandle
  # 1: Server name
  local $handle=$_[0];
  local $dest=$_[1];

  if ($#_!=1) {
    &UI::stderr("Invalid parameters passed to channel_open().\n");
    return 0;
  }
  if ($Yeemp::Config::offline{$dest}) {
    if (!$Yeemp::Config::rekeyed{$dest}) {
      return 0;
    } else {
      delete($Yeemp::Config::rekeyed{$dest});
      delete($Yeemp::Config::offline{$dest});
    }
  }
  if (!$openssl_initialized) {
    &initialize_openssl();
  }
  local $dest_ip = inet_aton($dest);
  if (!defined($dest_ip)) {
    UI::stderr("Unable to resolve address of $dest.\n");
    return 0;
  }
  if ($dest_ip =~ /^(....)$/s) {
    $dest_ip = $1;
  } else {
    UI::stderr("Dest IP not four bytes.  Odd.\n");
    return 0;
  }
  local $dest_serv_params;
  if (defined($Yeemp::Config::server_port)) {
    $dest_serv_params = sockaddr_in($Yeemp::Config::server_port, $dest_ip);
  } else {
    $dest_serv_params = sockaddr_in(51316, $dest_ip);
  }
  if (!socket  ($handle, &AF_INET(), &SOCK_STREAM(), 0)) {
    &UI::stderr("Error in socket call: $!\n");
    return 0;
  }
  if (!connect ($handle, $dest_serv_params)) {
    &UI::stderr("Connect to $dest failed: $!\n");
    &channel_close($handle);
    return 0;
  }
  select((select  ($handle), $| = 1)[0]);   # Eliminate STDIO buffering

  # Start SSL
  if (!($ctx{$handle} = Net::SSLeay::CTX_new())) {
    &UI::stderr("Failed to create SSL_CTX $!\n");
    &channel_close($handle);
    return 0;
  }
  local $rekeying;
  if (!Net::SSLeay::CTX_load_verify_locations($ctx{$handle},
    "$Yeemp::Config::certcache_dir/$dest", '')) {
    Net::SSLeay::print_errs();
    $need_cert{$dest}=1;
    delete($ctx{$handle});
    if (!($ctx{$handle} = Net::SSLeay::CTX_new())) {
      &UI::stderr("Failed to (re)create SSL_CTX $!\n");
      &channel_close($handle);
      return 0;
    }
    Net::SSLeay::CTX_set_verify($ctx{$handle}, &Net::SSLeay::VERIFY_PEER,
      \&verify_callback_newcert);
    $rekeying=1;
  } else {
    Net::SSLeay::CTX_set_verify($ctx{$handle}, &Net::SSLeay::VERIFY_PEER,
      \&verify_callback);
  }
  Net::SSLeay::CTX_set_options($ctx{$handle}, &Net::SSLeay::OP_ALL)
    and die_if_ssl_error("ssl ctx set options");
  die_if_ssl_error("SSL error in CTX_set_verify\n");
  $ssl{$handle} = Net::SSLeay::new($ctx{$handle}) or die_now("Failed to create SSL $!");
  local $target_domain=$dest;
  Net::SSLeay::set_fd($ssl{$handle}, fileno($handle));   # Must use fileno
  if (!Net::SSLeay::connect($ssl{$handle})) {
    &UI::stderr("Error $! connecting to SSL peer.\n");
    &channel_close($handle);
    return 0;
  }
  if (!$verify_callback_ok{$dest}) {
    &UI::stderr("Cert chain NOT verified.  Closing.\n");
    &channel_close($handle);
    return 0;
  }
  $channel_buf{$handle}='';
  $NEW_received{$handle}=0;
  return 1;
}

sub channel_write($$)
{
  if ($#_!=1) {
    &UI::stderr("$#_ args passed to channel_write.\n");
    return;
  }
  local $handle=$_[0];
  if (!defined($ssl{$handle})) {
    &UI::stderr("SSL channel undefined.\n");
    return;
  }
  $res = Net::SSLeay::ssl_write_all($ssl{$handle}, $_[1]);
  die_if_ssl_error("ssl write to $ssl{$handle}");
}

sub sanitize($)
{
  my $raw = $_[0];
  my $x=0;
  my $c;
  my $out='';
  while ($x<length($raw)) {
    $c = substr($raw,$x,1);
    if ((ord($c)<32) || (ord($c)>125)) {
      $out.='~'.sprintf('%02x',ord($c));
    } else {
     $out.=$c;
    }
    $x++;
  }
  return $out;
}

sub channel_slurp($)
{
  local $handle=$_[0];
  local $read = Net::SSLeay::read($ssl{$handle});
  die_if_ssl_error("Err on $handle/read\n");
  if (defined($read)) {
    if ($Yeemp::Config::debug>=2) {
      print STDERR "DEBUG2: Read ",sanitize($read),"\n";
    }
    $channel_buf{$handle}.=$read;
    if ($read eq '') {
      ## //FIXME: This may well close the channel wrongly on occasion!
      &UI::stderr("Blank read on SSL channel.  Closing.\n");
      channel_close($handle);
    } elsif ($channel_buf{$handle}=~/(^|\012)NEW:(|.*?)\012/) {
		  if ($NEW_received{$handle}<1000) {
        $NEW_received{$handle}++;
			} else {
			  &UI::stderr("Too many NEWs uncached.\n");
			}
      $channel_buf{$handle}=~s/^NEW:(|.*?)\012//s;
    }
  } else {
     &UI::debug("Undef read $handle\n");
  }
}

sub channel_readln($)
{
  if ($#_!=0) {
    &UI::stderr("Malformed call to channel_readln()\n");
    return;
  }
  local $handle=$_[0];
  if (!defined($ssl{$handle})) {
    &UI::stderr("Read on undef channel $handle.\n");
    return undef;
  }
  local $times=0;
  local $readstart=time();
  while ($channel_buf{$handle}!~/^(|.*?)\012/s) {
    local $read = Net::SSLeay::read($ssl{$handle});
    if (defined($read)) {
      if ($read ne '') {
        $channel_buf{$handle} .= $read;
      } else {
        $times++;
        if ($times>10) {
          if (time>$readstart+20) {
            ## //FIXME: This may well close the channel wrongly on occasion!
            &UI::stderr("Blank read on SSL channel.  Closing.\n");
            channel_close($handle);
            # Timeout the read.
            return undef;
          }
          # Slow down when read() wants to return '' over and over.
          select(undef,undef,undef,.1);
          $times=0;
        }
      }
    } else {
      &UI::stderr("Undef read; closing channel $handle.\n");
      $channel_buf{$handle}.="\012";
      &channel_close($handle);
    }
  }
  $channel_buf{$handle}=~/^(|.*?)\012/s;
  local $ret=$1;
  $channel_buf{$handle}=~s/^\Q$ret\E\012//s;
  die_if_ssl_error("ssl read of $handle failed.\n");
  return $ret;
}

sub channel_command($$)
{
  # Write $_[1] to the channel, followed by a newline.
  channel_write($_[0], "$_[1]\012");
  return &channel_read_response($_[0]);
}

sub channel_read_response($)
{
  local $notdone=1;
  local $line;
  while ($notdone) {
    $line=channel_readln($_[0]);
    if (defined($line)) {
      if ($line=~/^([012389]): (.*?|$)$/) {
        return ($1,': ',$2);
      } elsif ($line=~/^NEW:/) {
        $NEW_received{$_[0]}=1;
      } else {
        $notdone=0;
      }
    } else {
      $notdone=0;
    }
  }
  # If the server's response is incomprehensible, the original line will be in resp[1].
  return (undef,$line,undef);
}

sub channel_close($)
{
  local $handle=$_[0];
  if (defined($ssl{$handle})) {
    Net::SSLeay::free ($ssl{$handle});               # Tear down connection
  }
  if (defined($ctx{$handle})) {
    Net::SSLeay::CTX_free ($ctx{$handle});
  }
  if (defined(fileno($handle))) {
    shutdown $handle, 1;  # Half close --> No more output, sends EOF to server
    close $handle;
  }
  undef($ssl{$handle});
  undef($ctx{$handle});
  delete($NEW_received{$handle});
  undef($Yeemp::Client::channel_logged_in{$handle});
}

sub channel_is_open($)
{
  if (defined($_[0]) && defined(fileno($_[0]))) {
    return 1;
  }
  return 0;
}

sub channel_accept($)
{
  local $handle=$_[0];
  
  if (!$openssl_initialized) {
    &initialize_openssl();
  }
  if ($ctx{$handle} = Net::SSLeay::CTX_new) {
    Net::SSLeay::CTX_set_cipher_list($ctx{$handle},'ALL');
    if (&Net::SSLeay::set_cert_and_key($ctx{$handle}, "${Yeemp::Config::cert_path}yeempd.cert",
      "${Yeemp::Config::cert_path}yeempd.privkey")) {
      select((select($handle), $|=1)[0]);
      if ($ssl{$handle} = Net::SSLeay::new($ctx{$handle})) {
        Net::SSLeay::set_fd($ssl{$handle}, fileno($handle));
        Net::SSLeay::accept($ssl{$handle});
        $channel_buf{$handle}='';
        return 1;
      } else {
        &UI::stderr("Error creating ssl.\n");
      }
    } else {
      &UI::stderr("Error loading key.\n");
    }
  }
  return 0;
}

@ISA=qw(Exporter);
@EXPORT=qw(
    channel_close
	channel_command
	channel_is_open
	channel_open
	channel_readln
	channel_read_response
	channel_write
	channel_slurp
	);
